/*
 GBxCart RW - Wonderswan 8MB Flasher
 Version: 1.0
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 10/12/2022
 Last Modified: 13/01/2023
 License: CC-BY-NC-SA
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <signal.h>

#ifdef _WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#include "setup.h" // See defines, variables, constants, functions here

void sigintHandler(int sig_num) {
	set_mode('0');
	RS232_flushRX(cport_nr);
	gbx_cart_power_down();
	exit(0);
}

int main(int argc, char **argv) {
	signal(SIGINT, sigintHandler);
	
	printf("GBxCart RW - Wonderswan 8MB Flasher v1.0 by insideGadgets\n");
	printf("#########################################################\n");
	
	// Check arguments
	if (argc >= 2) {
		read_config();
		
		// Open COM port
		if (com_test_port() == 0) {
			printf("Device not connected and couldn't be auto detected. Please make sure the COM port isn't open elsewhere. If you are using Linux/Mac, make sure you use \"sudo\".\n");
			read_one_letter();
			return 1;
		}
		printf("Connected on COM port: %i at ", cport_nr+1);
		
		// Break out of any existing functions on ATmega
		set_mode('0');
		RS232_flushRX(cport_nr);
		
		// Get cartridge mode - Gameboy or Gameboy Advance
		cartridgeMode = request_value(CART_MODE);
		
		// Get PCB version
		gbxcartPcbVersion = request_value(READ_PCB_VERSION);
		xmas_wake_up();
		
		// Increase to 1.7Mbit baud for v1.4 PCB
		#ifdef _WIN32
		if ((gbxcartPcbVersion >= PCB_1_4 && gbxcartPcbVersion < GBXMAS) || gbxcartPcbVersion == MINI_1_1) {
			set_mode(USART_1_7M_SPEED);
			RS232_CloseComport(cport_nr);
			delay_ms(200);
			
			if (RS232_OpenComport(cport_nr, 1700000, "8N1") == 0) {
				printf("1.7M Baud\n");
			}
		}
		else {
			printf("1M Baud\n");
		}
		#else
			printf("1M Baud\n");
		#endif
		
		char pcbVersionString[20];
		if (gbxcartPcbVersion == PCB_1_0) {
			strncpy(pcbVersionString, "v1.0", 5);
		}
		else if (gbxcartPcbVersion == PCB_1_1) {
			strncpy(pcbVersionString, "v1.1/v1.2", 10);
		}
		else if (gbxcartPcbVersion == PCB_1_3) {
			strncpy(pcbVersionString, "v1.3", 5);
		}
		else if ((gbxcartPcbVersion >= PCB_1_4 && gbxcartPcbVersion < GBXMAS)) {
			strncpy(pcbVersionString, "v1.4", 5);
		}
		else if (gbxcartPcbVersion == MINI) {
			strncpy(pcbVersionString, "Mini", 5);
		}
		else if (gbxcartPcbVersion == MINI_1_1) {
			strncpy(pcbVersionString, "Mini v1.1", 10);
		}
		else if (gbxcartPcbVersion == GBXMAS) {
			strncpy(pcbVersionString, "XMAS", 5);
		}
		
		// Get firmware version
		gbxcartFirmwareVersion = request_value(READ_FIRMWARE_VERSION);
		printf("PCB version: %s, Firmware version: %i\n", pcbVersionString, gbxcartFirmwareVersion);
		
		if (gbxcartPcbVersion == PCB_1_0) {
			printf("\nPCB v1.0 is not supported for this function.");
			read_one_letter();
			return 1;
		}
		if (gbxcartFirmwareVersion == 0) {
			printf("\nFirmware L1 may be installed. In order to use this applications you will need to downgrade to R30.\n");
			read_one_letter();
			return 1;
		}
		if (gbxcartFirmwareVersion <= 8) {
			printf("Firmware R9 or higher is required for this functionality.\n");
			read_one_letter();
			return 1;
		}
		
		#ifdef TIMETEST
			// Timing test
			printf("Timing started\n");
			clock_t t;
			t = clock();
		#endif
		
		// Separate the file name
		char filenameOnly[100];
		char filename[254];
		strncpy(filename, argv[1], 253);
		
#ifdef _WIN32
		char *token = strtok(filename, "\\");
		while (token != NULL) {
			strncpy(filenameOnly, token, 99);
			token = strtok(NULL, "\\");
		}
#else
		char *token = strtok(filename, "/");
		while (token != NULL) {
			strncpy(filenameOnly, token, 99);
			token = strtok(NULL, "/");
		}
#endif
		
		
		printf("\nROM File selected: %s\n\n", filenameOnly);
		
		uint32_t readBytes = 0;
		long fileSize = 0;
		
		// Grab file size
		FILE *romFile = fopen(argv[1], "rb");
		if (romFile != NULL) {
			fseek(romFile, 0, SEEK_END);
			fileSize = ftell(romFile);
			fseek(romFile, 0, SEEK_SET);
		}
		else {
			printf("\n%s \nFile not found\n", argv[1]);
			read_one_letter();
			return 1;
		}
	
		// Check file size
		if (fileSize > 0x800000) {
			fclose(romFile);
			printf("\n%s \nFile size is larger than the available Flash cart space of 8 MByte\n", argv[1]);
			read_one_letter();
			return 1;
		}
	
		
		// PCB v1.3 or higher
		if (gbxcartPcbVersion == PCB_1_3 || (gbxcartPcbVersion >= PCB_1_4 && gbxcartPcbVersion < GBXMAS) || gbxcartPcbVersion == GBXMAS) {
			set_mode(VOLTAGE_3_3V);
			delay_ms(500);
			gbx_cart_power_up();
		}
		
		
		// WS Setup
		uint32_t wsBaseAddr = 0x20000;
		endAddr = 0x2FFFF;
		set_mode(WS_SETUP);
		delay_ms(250);
		
		ws_set_bank(WS_0C2, 127);
		delay_ms(20);
		
		// Check Flash ID
		set_number(0x20000, SET_START_ADDRESS);
		set_mode(WS_READ_ROM);
		com_read_bytes(READ_BUFFER, 64);
		com_read_stop(); // End read
		printf("Read ROM: 0x%X, 0x%X, 0x%X, 0x%X, 0x%X, 0x%X\n", readBuffer[0], readBuffer[1], readBuffer[2], readBuffer[3], readBuffer[4], readBuffer[5]);
		
		ws_flash_command(wsBaseAddr + 0xAAA, 0xAA);
		ws_flash_command(wsBaseAddr + 0x555, 0x55);
		ws_flash_command(wsBaseAddr + 0xAAA, 0x90);
		
		// Read ID
		set_number(0x20000, SET_START_ADDRESS);
		set_mode(WS_READ_ROM);
		com_read_bytes(READ_BUFFER, 64);
		com_read_stop(); // End read
		printf("Flash ID: 0x%X, 0x%X, 0x%X, 0x%X, 0x%X, 0x%X\n", readBuffer[0], readBuffer[1], readBuffer[2], readBuffer[3], readBuffer[4], readBuffer[5]);
		for (uint8_t r = 0; r < 8; r++) {
			flashID[r] = readBuffer[r];
		}
		ws_flash_command(wsBaseAddr + 0xAAA, 0xF0);
		
		
		// Check if Flash ID matches
		if (flashID[0] != 0xC2 && flashID[2] != 0xA8) {
			printf("*** Flash ID doesn't match expected value of 0xC2, 0xxx, 0xA8 ***\n");
			read_one_letter();
		}
		
		//read_one_letter();
		
		// Chip erase
		printf("\nErasing Flash");
		ws_flash_command(wsBaseAddr + 0xAAA, 0xAA);
		ws_flash_command(wsBaseAddr + 0x555, 0x55);
		ws_flash_command(wsBaseAddr + 0xAAA, 0x80);
		ws_flash_command(wsBaseAddr + 0xAAA, 0xAA);
		ws_flash_command(wsBaseAddr + 0x555, 0x55);
		ws_flash_command(wsBaseAddr + 0xAAA, 0x10);
		
		// Wait for first byte to be 0xFF
		ws_set_bank(WS_0C2, 0);
		ws_wait_for_flash_chip_erase_ff(1);
		
		
		printf("\n\nWriting to ROM (Flash cart) from %s\n", filenameOnly);
		printf("[             25%%             50%%             75%%            100%%]\n[");
		
		// Calculate banks needed from ROM file size
		romBanks = ceil((double) fileSize / (double) 65536);
		
		for (uint16_t bank = (128-romBanks); bank < 128; bank++) {
			ws_set_bank(WS_0C2, bank);
			delay_ms(5);
			
			currAddr = 0x20000;
			set_number(currAddr, SET_START_ADDRESS);
			delay_ms(5);
			
			// Read data
			while (currAddr < endAddr) {
				com_write_bytes_from_file(WS_FLASH_WRITE_64BYTE, romFile, 64);
				com_wait_for_ack();
				currAddr += 64;
				readBytes += 64;
				
				// Print progress
				print_progress_percent(readBytes, (romBanks * 65536) / 64);
				led_progress_percent(readBytes, (romBanks * 65536) / 28);
			}
		}
		printf("]");
		fclose(romFile);
		
		gbx_set_done_led();
		gbx_cart_power_down();
		
		printf("\n\nCompleted.\n");
	}
	else {
		printf("\n*** Please drag and drop your rom file to this program***\n");
	}
	
	read_one_letter();
	
	
	return 0;
}
